import requests
import uuid
import os
import textwrap
import json
import subprocess
import urllib
import jinja2
import pkgutil
from typing import List, Callable, Tuple
from vessel.pipeline import Step, Payload
from vessel.logging import logger
from vessel.version import LATEST_AGENT, LATEST_SENTINEL, LATEST_EVENT_ENGINE
from vessel.utilities import sanitizeClusterName


class DeployStep(Step):
  """
  Deploy generated yaml
  """

  def __init__(self, resource):
    super().__init__()
    self.resource = resource

  def run(self, payload:Payload, start_fn:Callable, end_fn:Callable, prompt_fn:Callable) -> Payload:
    start_fn("Deploying sentinel")
    if payload.token:
      yaml = os.path.expanduser(f"~/.daas/{payload.token}/{self.resource}.yaml")
    else:
      yaml = os.path.expanduser(f"~/.daas/{self.resource}.yaml")
    bashCommand = f"kubectl -n daas apply -f {yaml}"
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()
    if error:
      raise Exception(error)
    for out in output.decode('utf-8').split("\n"):
      end_fn(out)
    return payload

class GenerateYamlStep(Step):
  """
  Generate kube deployments for agent & sentinel
  """

  def __init__(self, resource, tag, more_vars=dict()):
    super().__init__()
    self.resource = resource
    self.tag = tag
    self.more_vars:dict = more_vars

  def _get_keys(self, payload:Payload) -> Tuple[str]:
    if payload.rsa:
      name = sanitizeClusterName(payload.cluster['result']['name'])
      return (name, payload.rsa, payload.vault_user, payload.vault_pwd, payload.distribution)
    else:
      with open(os.path.expanduser(f"~/.daas/{payload.token}/registration.json"), 'r') as f:
        registration = json.load(f)
      with open(os.path.expanduser(f"~/.daas/vault.json"), 'r') as f:
        vault = json.load(f)
      
      name = sanitizeClusterName(registration['cluster']['result']['name'])
      return (name, registration['rsa'], vault['public']['user'], vault['public']['pwd'], registration['distribution'])

  def run(self, payload:Payload, start_fn:Callable, end_fn:Callable, prompt_fn:Callable) -> Payload:
    
    
    if payload.token is not None:
      start_fn("Loading keys")
      path = os.path.expanduser(f"~/.daas/{payload.token}")
      (name, rsa, vault_user, vault_pwd, distribution) = self._get_keys(payload)

      tpl_vars = {
        "name": name, 
        "token": payload.token, 
        "rsa": rsa.replace("\n", "\\n"), 
        "vault_password": vault_pwd, 
        "distribution": distribution, 
        "tag": self.tag
      }
    else:
      path = os.path.expanduser(f"~/.daas")
      tpl_vars = {}

    tpl_vars.update(self.more_vars)
    

    
    start_fn(f"generating {self.resource} yaml")
    data = pkgutil.get_data(__name__, f"../resources/{self.resource}.yaml.jinja2")
    
    template = jinja2.Template(data.decode("utf-8"))
    # template = templateEnv.get_template(f"{self.resource}.yaml.jinja2")
    yaml = template.render(tpl_vars)
        
    with open(f"{path}/{self.resource}.yaml", 'w') as f:
      f.write(yaml)

    end_fn("OK")

    return payload



from typing import List, Callable
from vessel.pipeline import Step, Payload
from vessel.steps import VaultBaseStep
from vessel.logging import logger
import subprocess
import hvac
import base64
import os.path

class KubebenchStep(VaultBaseStep):
  """
  Unseal Vault Step
  """
  def get_rw_token(self, data):
    return base64.b64decode(data['private']['rw-token']).decode("utf-8")

  def __init__(self, vault_host, vault_info, yaml_dir):
    super().__init__(vault_host)
    self.vault_info = vault_info
    self.yaml_dir = yaml_dir
    #self.unlocked = unlocked

  def run(self, payload:Payload, start_fn:Callable, end_fn:Callable, prompt_fn:Callable) -> Payload:
    start_fn('Taking Vault root token')
    
    keys = self.get_vault_keys(payload, start_fn, prompt_fn)
    root_token = keys['root_token']
    self.vault.token = keys["root_token"]
    cluster_token = payload.token

    start_fn(root_token)
    start_fn(f"Cluster token is {cluster_token}. Taking k8s auth token from Vault")

    unlock_pub_secrets = self.vault.secrets.kv.v2.read_secret_version(
      path=cluster_token + "-daas-rw-ops",
      mount_point="daas_public"
    )

    pub_data = unlock_pub_secrets['data']['data']

    auth_token = pub_data['token']
    sa_name = pub_data['sa-name']
    distribution = pub_data['distribution']
    namespace = pub_data['namespace']

    start_fn(f"Cluster distribution of {cluster_token} is {distribution}")
    start_fn(f"Using the service account {sa_name} for installing kube-bench into the namespace {namespace}")

    if distribution == 'kubernetes':
      job_path = os.path.join(self.yaml_dir, 'k8s.yaml')
      bashCommand = f"kubectl -n {namespace} apply -f {job_path} --token={auth_token}"

    elif distribution == 'openshift':
      job_path = os.path.join(self.yaml_dir, 'ocp-3.11.yaml')
      bashCommand = f"oc -n {namespace} apply -f {job_path} --token={auth_token}"
  
    start_fn(f"Kube Bench yaml job path is {job_path}")
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()
    if error:
      raise Exception(error)
    for out in output.decode('utf-8').split("\n"):
      end_fn(out)

    end_fn(f"Job {job_path} applied!")



from pyfiglet import Figlet
import click
import time
import os
import sys
import json
import urllib
import urllib3
import json
import jinja2
from urllib.parse import urljoin
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from vessel.steps import *
from vessel.pipeline import Pipeline, Payload
from vessel.logging import logger
from vessel.version import VERSION, LATEST_AGENT, LATEST_SENTINEL
from vessel.utilities import sanitizeClusterName
import kubernetes
from kubernetes.client.rest import ApiException



def start(msg):
  click.echo(f"\n> {msg}...")

def end(msg):
  click.echo(f"[*] {msg}")

def prompt(msg, tag, **kwargs):
  return click.prompt(msg, type=str, **kwargs)

@click.group(invoke_without_command=True)
@click.option('--debug', is_flag=True, default=False, help="output debug log [False]")
@click.option('--version', is_flag=True, default=False,  help="Show version and exits")
@click.pass_context
def main(ctx, debug, version):
  """
  Vessel cli tool 
  """
  if version:
    print(f"Vessel cli tool: {VERSION}")
    sys.exit()

  if ctx.invoked_subcommand is None:
    print(ctx.get_help())
    sys.exit()

  click.clear()
  f = Figlet(font='standard')
  click.echo(f.renderText('Vessel cli tool'))
  ctx.ensure_object(dict)
  ctx.obj['DEBUG'] = debug

  if debug:
    logger.setLevel("DEBUG")
  
  # Init directories
  if not os.path.exists(os.path.expanduser("~/.daas")):
      os.mkdir( os.path.expanduser("~/.daas"))

@main.command()
@click.pass_context
@click.option('--vault', default='http://vault.local', help='Vault endpoint [http://vault.local]')
def init(ctx, vault):
  """
  Init vault
  """
  payload = Payload()
  pipeline = Pipeline(start_fn=start, end_fn=end, prompt_fn=prompt)
  pipeline.add(VaultInitStep(vault))
  try:
    # Run pipeline
    payload = pipeline.run(payload)    
  except Exception as e:
    logger.error(e)
    if ctx.obj['DEBUG']:
      raise e

@main.command()
@click.pass_context
@click.option('--vault', default='http://vault.local', help='Vault endpoint [http://vault.local]')
def unseal(ctx, vault):
  """
  Unseal vault
  """
  payload = Payload()
  pipeline = Pipeline(start_fn=start, end_fn=end, prompt_fn=prompt)
  pipeline.add(VaultUnsealStep(vault))
  try:
    # Run pipeline
    payload = pipeline.run(payload)    
  except Exception as e:
    logger.error(e)
    if ctx.obj['DEBUG']:
      raise e

@main.command()
@click.pass_context
@click.argument('token')
@click.option('--cluster-host', required=True, help="Hostname of the cluster to control")
@click.option('--cluster-ro', required=True, help="Cluster read-only service-account token")
@click.option('--cluster-rw', required=True, help="Cluster read-write service-account token")
@click.option('--vault', default='http://vault.local', help='Vault endpoint [http://vault.local]')
@click.option('--openshift', is_flag=True, default=False, help="Cluster is an Openshift distribution [False]")
@click.option('--init', is_flag=True, default=False, help="Initialize Vault [False]")
@click.option('--vessel-api', default="http://cloud-api.oc.corp.sourcesense.com/rpc", help="Vessel API RPC endpoint [http://cloud-api.oc.corp.sourcesense.com/rpc]")
def register(ctx, token, cluster_host, cluster_ro, cluster_rw, vault, openshift, init, vessel_api):
  """
  Register workstaion to Vessel with the given TOKEN
  """
  payload = Payload(token)
  pipeline = Pipeline(start_fn=start, end_fn=end, prompt_fn=prompt)
  if init:
    pipeline.add(VaultInitStep(vault))
  pipeline.add(VaultSaveSecretsStep(vault, cluster_host, cluster_ro, cluster_rw, openshift))
  pipeline.add(GenerateKeysStep())
  pipeline.add(RegisterStep(vessel_api))  

  try:
    # Run pipeline
    payload = pipeline.run(payload)    
  except Exception as e:
    logger.error(e)
    if ctx.obj['DEBUG']:
      raise e
  

@main.command()
@click.pass_context
@click.argument('token')
@click.option('--sentinel', help="Generates sentinel at given tag [None]")
@click.option('--agent', help="Generates agent yaml at given tag [None]")
@click.option('--event', help="Generates event-engine yaml at given tag [None]")
@click.option('--apply', is_flag=True, default=False, help="run kubectl apply on generated deployments [False]")

def deployment(ctx, token, sentinel, agent, event, apply):
  """
  Generated and Deploy agent, sentinel and event-engine deployments to internal kubernetes for given TOKEN
  """
  click.echo('Debug is %s' % (ctx.obj['DEBUG'] and 'on' or 'off'))
  payload = Payload(token)
  pipeline = Pipeline(start_fn=start, end_fn=end, prompt_fn=prompt)
  
  if sentinel:
    pipeline.add(GenerateYamlStep("sentinel", sentinel if sentinel != "latest" else LATEST_SENTINEL))
    if apply:
      pipeline.add(DeployStep("sentinel"))

  if agent:
    pipeline.add(GenerateYamlStep("agent", agent if agent != "latest" else LATEST_AGENT))
    if apply:
      pipeline.add(DeployStep("agent"))

  if event:
    pipeline.add(GenerateYamlStep("event-engine", event if event != "latest" else LATEST_EVENT_ENGINE))
    if apply:
      pipeline.add(DeployStep("event-engine"))

  try:
    # Run pipeline
    payload = pipeline.run(payload)    
  except Exception as e:
    logger.error(e)
    if ctx.obj['DEBUG']:
      raise e


@main.command()
@click.pass_context
@click.argument('svc', nargs=-1, required=True)
@click.option('--host', default="vessel-workstation.local", help="Hostname used by the generated ingress")
@click.option('--apply', is_flag=True, default=False, help="run kubectl apply on generated Ingress [False]")
def alertmanager(ctx, svc, host, apply):
  """
  Gnerate and apply an Ingress configuration for the event-engine http interface specified by the svc name passed, and output a uitable alertmanager configuration for prometheus
  """
  payload = Payload()
  pipeline = Pipeline(start_fn=start, end_fn=end, prompt_fn=prompt)
  pipeline.add(GenerateYamlStep("event-engine-ingress", None, {
    "svc": svc,
    "host": host
  }))
  if apply:
      pipeline.add(DeployStep("event-engine-ingress"))
  try:
    # Run pipeline
    payload = pipeline.run(payload)
    start("Genrating prometheus config file to stdout")
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "resources", "alertmanager.yaml.jinja2"), "r") as tpl:
      template = jinja2.Template(tpl.read())
      out = template.render(svc=svc, host=host)
      
      print(f"\n\n{out}")
    # output alertmanager config
    # tempalte_dir = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "resources")
    # templateEnv = jinja2.Environment(loader=jinja2.FileSystemLoader(searchpath=tempalte_dir))
    
    # template = templateEnv.get_template(f"{self.resource}.yaml.jinja2")
    # yaml = template.render(tpl_vars)
  except Exception as e:
    logger.error(e)
    if ctx.obj['DEBUG']:
      raise e

@main.command()
@click.pass_context
@click.argument('token')
@click.option('--agent-tag', default=LATEST_AGENT, help="Set image tag for agent deployment")
@click.option('--sentinel-tag', default=LATEST_SENTINEL, help="Set image tag for sentinel deployment")
@click.option('--vault', default='http://vault.local', help='Vault endpoint [http://vault.local]')
def update(ctx, token, vault, sentinel_tag, agent_tag):
  """
  Updates agent and sentinel deployments
  """
  
  payload = Payload(token)
  pipeline = Pipeline(start_fn=start, end_fn=end, prompt_fn=prompt)
  pipeline.add(UpdateStep(agent_tag, sentinel_tag))

  try:
    # Run pipeline
    payload = pipeline.run(payload)    

  except Exception as e:
    logger.error(e)
    if ctx.obj['DEBUG']:
      raise e

@main.command()
@click.pass_context
@click.argument('token')
@click.option('--vault', default='http://vault.local', help='Vault endpoint [http://vault.local]')
@click.option('--vault-info', default='/home/rke/.daas/vault.json', help='Local file for Vault access information')
@click.option('--yaml-dir', default='/home/rke/k8s_yaml/kube-bench', help='Folder containing k8s and openshift yaml files')
def kubebench(ctx, token, vault, vault_info, yaml_dir):
  payload = Payload(token)
  pipeline = Pipeline(start_fn=start, end_fn=end, prompt_fn=prompt)
  pipeline.add(KubebenchStep(vault, vault_info, yaml_dir))

  try:
    # Run pipeline
    payload = pipeline.run(payload)    
  except Exception as e:
    logger.error(e)
    if ctx.obj['DEBUG']:
      raise e


@main.command()
def completion():
  """
  Install completion script
  """
  with open("./vessel-cli.bash", "w+") as f:
    f.write("#/usr/bin/env bash\ncomplete -W $(ls ~/.daas/ | grep -v json) vessel-cli")
  click.echo("saved completion script in ./vessel-cli.bash\n\nto activete:\n\necho source ./vessel-cli.bash >> ~/.bash_profile")